import platform
from typing import Any
import pytest
from demagnetize.bencode import bencode, partial_unbencode, unbencode
from demagnetize.errors import UnbencodeError


@pytest.mark.parametrize(
    "blob,data",
    [
        (b"4:spam", b"spam"),
        (b"i3e", 3),
        (b"l4:spam4:eggse", [b"spam", b"eggs"]),
        (b"d3:cow3:moo4:spam4:eggse", {b"cow": b"moo", b"spam": b"eggs"}),
        (b"d4:data4:\x00\x01\x02\x03e", {b"data": b"\x00\x01\x02\x03"}),
        (b"i0e", 0),
        (b"i-1e", -1),
        (b"i-10e", -10),
        (b"i35e", 35),
        (b"0:", b""),
        (b"le", []),
        (b"de", {}),
        (b"d8:msg_typei0e5:piecei0ee", {b"msg_type": 0, b"piece": 0}),
        (
            b"d1:md11:ut_metadatai3ee13:metadata_sizei31235ee",
            {b"m": {b"ut_metadata": 3}, b"metadata_size": 31235},
        ),
    ],
)
def test_bencode(blob: bytes, data: Any) -> None:
    assert unbencode(blob) == data
    assert bencode(data) == blob


@pytest.mark.parametrize(
    "blob",
    [
        b"i-0e",
        b"i00e",
        b"i04e",
        b"04:spam",
        b"-4:spam",
        b"-0:",
        b"24:short",
        b"4:longextra",
        b"l",
        b"q",
        b"d",
        b"di32e6:stringe",
        b"d6:bananai1e5:applei2e",
        b"i3.14e",
        b"i12-e",
        b"i 12e",
        b"i12 e",
        b"i12:",
        b"5eapple",
        pytest.param(
            b"l" * 1234 + b"e" * 1234,
            marks=pytest.mark.skipif(
                platform.python_implementation() == "PyPy",
                reason="Recursion depth is weird on PyPy",
            ),
        ),
    ],
)
def test_unbencode_error(blob: bytes) -> None:
    with pytest.raises(UnbencodeError):
        unbencode(blob)


@pytest.mark.parametrize(
    "blob,data,trailing",
    [
        (
            b"d8:msg_typei1e5:piecei0e10:total_sizei3425eeabcdefg",
            {b"msg_type": 1, b"piece": 0, b"total_size": 3425},
            b"abcdefg",
        ),
    ],
)
def test_partial_unbencode(blob: bytes, data: Any, trailing: bytes) -> None:
    assert partial_unbencode(blob) == (data, trailing)



from __future__ import annotations
from collections.abc import AsyncGenerator
from contextlib import aclosing
from typing import TYPE_CHECKING
from anyio import (
    CapacityLimiter,
    EndOfStream,
    Event,
    create_memory_object_stream,
    create_task_group,
)
from anyio.abc import TaskGroup
from anyio.streams.memory import MemoryObjectSendStream
import attr
from torf import Magnet
from .consts import PEERS_PER_MAGNET_LIMIT
from .errors import DemagnetizeError, PeerError
from .peer import Peer, PeerAddress
from .trackers import Tracker
from .util import TRACE, InfoHash, log

if TYPE_CHECKING:
    from .core import Demagnetizer


@attr.define
class TorrentSession:
    app: Demagnetizer
    magnet: Magnet
    info_hash: InfoHash = attr.field(init=False)
    peer_limit: CapacityLimiter = attr.field(init=False)
    peers_seen: set[PeerAddress] = attr.Factory(set)

    def __attrs_post_init__(self) -> None:
        # torf only accepts magnet links with valid info hashes, so this
        # shouldn't fail:
        self.info_hash = InfoHash.from_string(self.magnet.infohash)
        self.peer_limit = CapacityLimiter(PEERS_PER_MAGNET_LIMIT)

    async def get_info(self) -> dict:
        if not self.magnet.tr:
            raise DemagnetizeError(
                f"Cannot fetch info for info hash {self.info_hash}: No trackers"
                " in magnet link"
            )
        if self.magnet.dn is not None:
            display = f" ({self.magnet.dn})"
        else:
            display = ""
        log.info("Fetching info for info hash %s%s", self.info_hash, display)
        async with create_task_group() as tg:
            peer_aiter = self.get_all_peers(tg)
            info_sender, info_receiver = create_memory_object_stream(0, dict)
            tg.start_soon(self._peer_pipe, peer_aiter, info_sender, tg)
            async with info_receiver:
                try:
                    md = await info_receiver.receive()
                except EndOfStream:
                    raise DemagnetizeError(f"Failed to fetch info for {self.info_hash}")
            tg.cancel_scope.cancel()
            return md

    async def get_all_peers(self, task_group: TaskGroup) -> AsyncGenerator[Peer, None]:
        sender, receiver = create_memory_object_stream()
        async with sender:
            for url in self.magnet.tr:
                try:
                    tracker = Tracker.from_url(url)
                except ValueError as e:
                    log.warning("%s: Invalid tracker URL: %s", url, e)
                else:
                    task_group.start_soon(
                        tracker.get_peers, self.app, self.info_hash, sender.clone()
                    )
        async with receiver:
            async for p in receiver:
                if (addr := p.address) not in self.peers_seen:
                    self.peers_seen.add(addr)
                    yield p
                else:
                    log.log(TRACE, "%s returned by multiple trackers; skipping", p)

    async def _peer_pipe(
        self,
        peer_aiter: AsyncGenerator[Peer, None],
        info_sender: MemoryObjectSendStream[dict],
        task_group: TaskGroup,
    ) -> None:
        info_fetched = Event()
        async with aclosing(peer_aiter), info_sender:
            async for peer in peer_aiter:
                task_group.start_soon(
                    self._peer_task, peer, info_sender.clone(), info_fetched
                )

    async def _peer_task(
        self, peer: Peer, info_sender: MemoryObjectSendStream[dict], info_fetched: Event
    ) -> None:
        async with self.peer_limit, info_sender:
            if not info_fetched.is_set():
                try:
                    info = await peer.get_info(self.app, self.info_hash)
                except PeerError as e:
                    log.warning(
                        "Error getting info for %s from %s: %s",
                        self.info_hash,
                        peer,
                        e.msg,
                    )
                else:
                    log.info("Received info for %s from %s", self.info_hash, peer)
                    info_fetched.set()
                    await info_sender.send(info)



from __future__ import annotations
from typing import TYPE_CHECKING, NoReturn, Optional
from anyio import (
    BrokenResourceError,
    ClosedResourceError,
    EndOfStream,
    IncompleteRead,
    connect_tcp,
    fail_after,
)
from anyio.abc import ObjectStream, SocketStream
from anyio.streams.buffered import BufferedByteReceiveStream
import attr
from .extensions import BEP9MsgType, BEP10Extension, BEP10Registry, Extension
from .messages import (
    AllowedFast,
    AnyMessage,
    BEP9Message,
    Bitfield,
    EmptyMessage,
    ExtendedHandshake,
    Handshake,
    Have,
    HaveNone,
    Piece,
    Suggest,
    decode_message,
    encode_message,
)
from ..bencode import unbencode
from ..consts import CLIENT, MAX_PEER_MSG_LEN, PEER_HANDSHAKE_TIMEOUT, UT_METADATA
from ..errors import PeerError, UnbencodeError
from ..util import TRACE, InfoHash, InfoPiecer, log

if TYPE_CHECKING:
    from ..core import Demagnetizer


SUPPORTED_EXTENSIONS = {Extension.BEP10_EXTENSIONS, Extension.FAST}

LOCAL_BEP10_REGISTRY = BEP10Registry.from_dict(
    {
        BEP10Extension.METADATA: UT_METADATA,
    }
)

IGNORED_MESSAGES = (
    EmptyMessage,
    Have,
    Bitfield,
    Piece,
    AllowedFast,
    Suggest,
    # It's valid for a peer to send an extended handshake more than once, and
    # it's valid for us to ignore subsequent handshakes, so that's what we'll
    # do.
    ExtendedHandshake,
)

PeerAddress = tuple[str, int]


@attr.define
class Peer:
    host: str
    port: int
    id: Optional[bytes] = None  # noqa: A003

    def __str__(self) -> str:
        if ":" in self.host:
            addr = f"[{self.host}]:{self.port}"
        else:
            addr = f"{self.host}:{self.port}"
        return f"<Peer {addr}>"

    @property
    def address(self) -> PeerAddress:
        return (self.host, self.port)

    async def get_info(self, app: Demagnetizer, info_hash: InfoHash) -> dict:
        log.info("Requesting info for %s from %s", info_hash, self)
        try:
            try:
                with fail_after(PEER_HANDSHAKE_TIMEOUT):
                    conn = await self.connect(app, info_hash)
            except TimeoutError:
                raise PeerError(
                    peer=self,
                    info_hash=info_hash,
                    msg="Could not connect to peer in time",
                )
            async with conn:
                return await get_metadata_info(conn)
        except OSError as e:
            raise PeerError(
                peer=self,
                info_hash=info_hash,
                msg=f"Communication error: {type(e).__name__}: {e}",
            )

    async def connect(self, app: Demagnetizer, info_hash: InfoHash) -> PeerConnection:
        log.debug("Connecting to %s", self)
        s = await connect_tcp(self.host, self.port)
        log.log(TRACE, "Connected to %s", self)
        conn = PeerConnection(peer=self, app=app, socket=s, info_hash=info_hash)
        try:
            await conn.handshake()
        except BaseException:
            await conn.aclose()
            raise
        return conn


@attr.define
class PeerConnection(ObjectStream[AnyMessage]):
    peer: Peer
    app: Demagnetizer
    socket: SocketStream
    info_hash: InfoHash
    readstream: BufferedByteReceiveStream = attr.field(init=False)
    extensions: set[Extension] = attr.Factory(set)
    remote_bep10_registry: BEP10Registry = attr.Factory(BEP10Registry)

    def __attrs_post_init__(self) -> None:
        self.readstream = BufferedByteReceiveStream(self.socket)

    async def aclose(self) -> None:
        await self.socket.aclose()

    async def send(self, msg: AnyMessage) -> None:
        log.log(TRACE, "Sending to %s: %s", self.peer, msg)
        try:
            await self.socket.send(encode_message(msg, self.remote_bep10_registry))
        except (BrokenResourceError, ClosedResourceError):
            self.error("Peer closed the connection early")

    async def send_eof(self) -> None:
        await self.socket.send_eof()

    async def read(self, length: int) -> bytes:
        try:
            return await self.readstream.receive_exactly(length)
        except (EndOfStream, IncompleteRead, BrokenResourceError, ClosedResourceError):
            self.error("Peer closed the connection early")

    async def handshake(self) -> None:
        log.log(TRACE, "Sending handshake to %s", self.peer)
        try:
            await self.socket.send(
                bytes(
                    Handshake(
                        extensions=set(SUPPORTED_EXTENSIONS),  # set() for mypy
                        info_hash=self.info_hash,
                        peer_id=self.app.peer_id,
                    )
                )
            )
        except (BrokenResourceError, ClosedResourceError):
            self.error("Peer closed the connection early")
        r = await self.read(Handshake.LENGTH)
        try:
            hs = Handshake.parse(r)
        except ValueError as e:
            log.log(TRACE, "Bad handshake from %s: %r", self.peer, r)
            self.error(f"Peer sent bad handshake: {e}")
        log.log(TRACE, "%s sent %s", self.peer, hs)
        if hs.info_hash != self.info_hash:
            self.error(f"Peer replied with wrong info hash (got {hs.info_hash})")
        self.extensions = SUPPORTED_EXTENSIONS & hs.extensions
        if Extension.BEP10_EXTENSIONS in self.extensions:
            await self.send(
                ExtendedHandshake.make(extensions=LOCAL_BEP10_REGISTRY, client=CLIENT)
            )
        else:
            self.error("Peer does not support BEP 10 extensions")
        if Extension.FAST in self.extensions:
            await self.send(HaveNone())

    def error(self, msg: str) -> NoReturn:
        raise PeerError(peer=self.peer, info_hash=self.info_hash, msg=msg)

    async def receive(self) -> AnyMessage:
        while True:
            blen = await self.read(4)
            length = int.from_bytes(blen, "big")
            if length > MAX_PEER_MSG_LEN:
                self.error(
                    f"Peer tried to send overly large message of {length}"
                    " bytes; not trusting"
                )
            if length == 0:
                log.log(TRACE, "%s sent keepalive", self.peer)
            else:
                payload = await self.read(length)
                try:
                    msg = decode_message(blen + payload, LOCAL_BEP10_REGISTRY)
                except ValueError as e:
                    log.log(TRACE, "Bad message from %s: %r", self.peer, payload)
                    self.error(f"Peer sent invalid message: {e}")
                else:
                    log.log(TRACE, "%s sent message: %s", self.peer, msg)
                    return msg


async def get_metadata_info(conn: PeerConnection) -> dict:
    # Unlike a normal torrent, we expect to get the entire info from a single
    # peer and error if it can't give it to us (because peers should only be
    # sending any info if they've checked the whole thing, and if they can't
    # send it all, why should we trust them?)
    ### TODO: Put a timeout on this:
    async for msg in conn:
        if isinstance(msg, ExtendedHandshake):
            handshake = msg
            break
        elif not isinstance(msg, IGNORED_MESSAGES):
            conn.error(f"Peer sent unexpected message: {msg}")
    conn.remote_bep10_registry = handshake.extensions
    if BEP10Extension.METADATA not in conn.remote_bep10_registry:
        conn.error("Peer does not support metadata transfer")
    if handshake.metadata_size is None:
        conn.error("Peer did not report info size in extended handshake")
    log.log(
        TRACE, "%s declares info size as %d bytes", conn.peer, handshake.metadata_size
    )
    info_piecer = InfoPiecer(handshake.metadata_size)
    for i in range(info_piecer.piece_qty):
        log.debug(
            "Sending request to %s for info piece %d/%d",
            conn.peer,
            i,
            info_piecer.piece_qty,
        )
        await conn.send(BEP9Message(msg_type=BEP9MsgType.REQUEST, piece=i))
        async for msg in conn:
            if isinstance(msg, BEP9Message):
                if msg.msg_type == BEP9MsgType.DATA:
                    if msg.piece != i:
                        conn.error(
                            f"received data for info piece {msg.piece}, which"
                            " we did not request"
                        )
                    elif (
                        msg.total_size is not None
                        and msg.total_size != info_piecer.total_size
                    ):
                        conn.error(
                            "'total_size' in info data message"
                            f" ({msg.total_size}) differs from previous value"
                            f" ({info_piecer.total_size})"
                        )
                    log.debug("%s sent info piece %d", conn.peer, msg.piece)
                    try:
                        info_piecer.add_piece(msg.payload)
                    except ValueError as e:
                        conn.error(f"bad info piece: {e}")
                    break
                elif msg.msg_type == BEP9MsgType.REJECT:
                    conn.error(f"Peer rejected request for info piece {msg.piece}")
                elif msg.msg_type == BEP9MsgType.REQUEST:
                    log.log(
                        TRACE,
                        "%s sent request for info piece %d; rejecting",
                        conn.peer,
                        msg.piece,
                    )
                    await conn.send(
                        BEP9Message(msg_type=BEP9MsgType.REJECT, piece=msg.piece)
                    )
                else:
                    log.log(
                        TRACE,
                        "%s sent ut_metadata message with unknown msg_type %d;"
                        " ignoring",
                        conn.peer,
                        msg.msg_type,
                    )
            elif not isinstance(msg, IGNORED_MESSAGES):
                conn.error(f"Peer sent unexpected message: {msg}")
    log.debug("All info pieces received from %s; validating ...", conn.peer)
    if (good_dgst := conn.info_hash.as_hex) != (dgst := info_piecer.get_digest()):
        conn.error(
            f"Received info with invalid digest; expected {good_dgst}, got {dgst}"
        )
    data = info_piecer.get_data()
    try:
        info = unbencode(data)
    except UnbencodeError as e:
        conn.error(f"Received invalid bencoded data as info: {e}")
    if not isinstance(info, dict):
        conn.error("Received bencoded non-dict as info")
    return info



from __future__ import annotations
from abc import ABC, abstractmethod
from enum import Enum
from ipaddress import AddressValueError, IPv4Address, IPv6Address
import struct
from typing import TYPE_CHECKING, ClassVar
from anyio import fail_after
from anyio.abc import AsyncResource
from anyio.streams.memory import MemoryObjectSendStream
import attr
from yarl import URL
from ..consts import LEFT, TRACKER_STOP_TIMEOUT, TRACKER_TIMEOUT
from ..errors import TrackerError, TrackerFailure
from ..peer import Peer
from ..util import TRACE, InfoHash, log

if TYPE_CHECKING:
    from ..core import Demagnetizer


@attr.define
class Tracker(ABC):
    SCHEMES: ClassVar[list[str]]
    url: URL

    def __str__(self) -> str:
        return f"<Tracker {self.url}>"

    @classmethod
    def from_url(cls, url: str) -> Tracker:
        u = URL(url)  # Raises ValueError on failure
        for klass in Tracker.__subclasses__():
            if u.scheme in klass.SCHEMES:
                return klass(url=u)  # type: ignore[abstract]
        raise ValueError(f"Unsupported tracker URL scheme {u.scheme!r}")

    async def get_peers(
        self,
        app: Demagnetizer,
        info_hash: InfoHash,
        sender: MemoryObjectSendStream[Peer],
    ) -> None:
        log.info("Requesting peers for %s from %s", info_hash, self)
        try:
            with fail_after(TRACKER_TIMEOUT):
                async with sender, await self.connect(app) as conn:
                    log.log(
                        TRACE,
                        "Sending 'started' announcement to %s for %s",
                        self,
                        info_hash,
                    )
                    peers = (
                        await conn.announce(info_hash, AnnounceEvent.STARTED)
                    ).peers
                    log.info("%s returned %d peers for %s", self, len(peers), info_hash)
                    log.debug(
                        "%s returned peers for %s: %s",
                        self,
                        info_hash,
                        ", ".join(map(str, peers)) or "<none>",
                    )
                    for p in peers:
                        await sender.send(p)
                    with fail_after(TRACKER_STOP_TIMEOUT, shield=True):
                        log.log(
                            TRACE,
                            "Sending 'stopped' announcement to %s for %s",
                            self,
                            info_hash,
                        )
                        await conn.announce(info_hash, AnnounceEvent.STOPPED)
        except TrackerError as e:
            log.warning(
                "Error getting peers for %s from %s: %s", info_hash, self, e.msg
            )
        except TrackerFailure as e:
            log.warning(
                "%s replied to announcement for %s with failure message: %s",
                self,
                info_hash,
                e,
            )
        except TimeoutError:
            log.warning("%s did not respond in time", self)
        except OSError as e:
            log.warning(
                "Error communicating with %s: %s: %s", self, type(e).__name__, e
            )

    @abstractmethod
    async def connect(self, app: Demagnetizer) -> TrackerSession:
        ...


class TrackerSession(AsyncResource):
    @abstractmethod
    async def announce(
        self,
        info_hash: InfoHash,
        event: AnnounceEvent,
        downloaded: int = 0,
        uploaded: int = 0,
        left: int = LEFT,
    ) -> AnnounceResponse:
        ...


class AnnounceEvent(Enum):
    ANNOUNCE = ("", 0)
    COMPLETED = ("completed", 1)
    STARTED = ("started", 2)
    STOPPED = ("stopped", 3)

    def __init__(self, http_value: str, udp_value: int) -> None:
        self.http_value = http_value
        self.udp_value = udp_value


@attr.define
class AnnounceResponse:
    interval: int
    peers: list[Peer]


def unpack_peers(data: bytes) -> list[Peer]:
    peers: list[Peer] = []
    try:
        for (ipnum, port) in struct.iter_unpack("!IH", data):
            ip = str(IPv4Address(ipnum))
            peers.append(Peer(host=ip, port=port))
    except (struct.error, AddressValueError):
        raise ValueError("invalid 'peers' list")
    return peers


def unpack_peers6(data: bytes) -> list[Peer]:
    peers6: list[Peer] = []
    try:
        for (*ipbytes, port) in struct.iter_unpack("!16cH", data):
            ip = str(IPv6Address(b"".join(ipbytes)))
            peers6.append(Peer(host=ip, port=port))
    except (struct.error, AddressValueError):
        raise ValueError("invalid 'peers6' list")
    return peers6



from __future__ import annotations
from typing import Any, Optional
import attr
from .errors import UnbencodeError


def bencode(obj: Any) -> bytes:
    if isinstance(obj, bytes):
        return b"%d:%b" % (len(obj), obj)
    elif isinstance(obj, int):
        return b"i%de" % (obj,)
    elif isinstance(obj, list):
        return b"l" + b"".join(bencode(o) for o in obj) + b"e"
    elif isinstance(obj, dict):
        s = b"d"
        for key in sorted(obj.keys()):
            if not isinstance(key, bytes):
                raise TypeError(f"Cannot bencode {type(key).__name__} dict keys")
            s += bencode(key) + bencode(obj[key])
        return s + b"e"
    else:
        raise TypeError(f"Cannot bencode {type(obj).__name__}")


@attr.define
class Unbencoder:
    buff: bytes
    index: int = 0

    def getchar(self) -> bytes:
        if self.index < len(self.buff):
            b = self.buff[self.index : self.index + 1]
            self.index += 1
            return b
        else:
            raise UnbencodeError("Short input")

    def peek(self) -> bytes:
        return self.buff[self.index : self.index + 1]

    def read_bytes(self, length: int) -> bytes:
        if self.index + length <= len(self.buff):
            blob = self.buff[self.index : self.index + length]
            self.index += length
            return blob
        else:
            raise UnbencodeError("Short input")

    def read_int(self, stop: bytes) -> int:
        digits: list[bytes] = []
        while (d := self.getchar()) != stop:
            if not d.isdigit() and not (d == b"-" and not digits):
                raise UnbencodeError("Non-digit in integer")
            digits.append(d)
        num = b"".join(digits).decode("us-ascii")
        if not num or (num[0] == "0" and len(digits) > 1) or (num[0:2] == "-0"):
            raise UnbencodeError("Invalid bencoded integer")
        return int(num, 10)

    def get_trailing(self) -> bytes:
        return self.buff[self.index :]

    def decode_next(self) -> Any:
        c = self.getchar()
        if c == b"d":
            bdict: dict[bytes, Any] = {}
            prev_key: Optional[bytes] = None
            while self.peek() != b"e":
                key = self.decode_next()
                if not isinstance(key, bytes):
                    raise UnbencodeError("Non-bytes key in dict")
                elif prev_key is not None and key <= prev_key:
                    raise UnbencodeError("Dict keys not in sorted order")
                value = self.decode_next()
                bdict[key] = value
                prev_key = key
            self.getchar()
            return bdict
        elif c == b"l":
            blist: list = []
            while self.peek() != b"e":
                blist.append(self.decode_next())
            self.getchar()
            return blist
        elif c == b"i":
            return self.read_int(b"e")
        elif c.isdigit():
            self.index -= 1
            length = self.read_int(b":")
            return self.read_bytes(length)
        else:
            raise UnbencodeError("Invalid byte in input")


def unbencode(blob: bytes) -> Any:
    (value, trailing) = partial_unbencode(blob)
    if trailing:
        raise UnbencodeError("Input contains trailing bytes")
    return value


def partial_unbencode(blob: bytes) -> tuple[Any, bytes]:
    decoder = Unbencoder(blob)
    try:
        value = decoder.decode_next()
    except RecursionError:
        raise UnbencodeError("Too many nested structures")
    return (value, decoder.get_trailing())



# NOTE: Avoid using slotted classes here, as combining them with attrs and
# `__subclasses__` can lead to Heisenbugs depending on when garbage collection
# happens.
from __future__ import annotations
from abc import ABC, abstractmethod
from collections import deque
from functools import reduce
from operator import or_
import struct
from typing import Any, ClassVar, Optional
import attr
from .extensions import BEP9MsgType, BEP10Extension, BEP10Registry, Extension
from ..bencode import bencode, partial_unbencode, unbencode
from ..errors import UnbencodeError
from ..util import InfoHash, get_string, get_typed_value


@attr.define(slots=False)
class Handshake:
    HEADER: ClassVar[bytes] = b"\x13BitTorrent protocol"
    LENGTH: ClassVar[int] = 20 + 8 + 20 + 20

    extensions: set[int]
    info_hash: InfoHash
    peer_id: bytes

    def __str__(self) -> str:
        extensions = ", ".join(self.extension_names) or "<none>"
        return f"handshake; extensions: {extensions}; peer_id: {self.peer_id!r}"

    def __bytes__(self) -> bytes:
        return (
            self.HEADER
            + reduce(or_, [1 << i for i in self.extensions], 0).to_bytes(8, "big")
            + bytes(self.info_hash)
            + (self.peer_id + b"\0" * 20)[:20]
        )

    @classmethod
    def parse(cls, blob: bytes) -> Handshake:
        if len(blob) != cls.LENGTH:
            raise ValueError(
                f"handshake wrong length; got {len(blob)} bytes, expected {cls.LENGTH}"
            )
        if blob[: len(cls.HEADER)] != cls.HEADER:
            raise ValueError("handshake had invalid protocol declaration")
        offset = len(cls.HEADER)
        exts = int.from_bytes(blob[offset : offset + 8], "big")
        extensions = {i for i in range(64) if exts & (1 << i)}
        offset += 8
        info_hash = InfoHash.from_bytes(blob[offset : offset + 20])
        offset += 20
        peer_id = blob[offset:]
        return cls(extensions=extensions, info_hash=info_hash, peer_id=peer_id)

    @property
    def extension_names(self) -> list[str]:
        extnames: list[str] = []
        for ext in sorted(self.extensions):
            try:
                extnames.append(Extension(ext).name)
            except ValueError:
                extnames.append(str(ext))
        return extnames


@attr.define(slots=False)
class Message(ABC):
    TYPE: ClassVar[int]

    def __bytes__(self) -> bytes:
        payload = self.to_payload()
        length = 1 + len(payload)
        return length.to_bytes(4, "big") + bytes([self.TYPE]) + payload

    @classmethod
    def parse(cls, blob: bytes) -> Message:
        # length = blob[:4]
        mtype = blob[4]
        payload = blob[5:]
        klasses = deque(Message.__subclasses__())
        while klasses:
            klass = klasses.popleft()
            if hasattr(klass, "TYPE"):
                if klass.TYPE == mtype:
                    return klass.from_payload(payload)
            else:
                klasses.extend(klass.__subclasses__())
        raise ValueError(f"Unknown message type: {mtype}")

    @classmethod
    @abstractmethod
    def from_payload(cls, payload: bytes) -> Message:
        ...

    @abstractmethod
    def to_payload(self) -> bytes:
        ...


@attr.define(slots=False)  # To make the empty messages into useful classes
class EmptyMessage(Message):
    @classmethod
    def from_payload(cls, _payload: bytes) -> EmptyMessage:
        ### TODO: Do something if payload is not empty?
        return cls()

    def to_payload(self) -> bytes:
        return b""


class Choke(EmptyMessage):
    TYPE = 0

    def __str__(self) -> str:
        return "choke"


class Unchoke(EmptyMessage):
    TYPE = 1

    def __str__(self) -> str:
        return "unchoke"


class Interested(EmptyMessage):
    TYPE = 2

    def __str__(self) -> str:
        return "interested"


class NotInterested(EmptyMessage):
    TYPE = 3

    def __str__(self) -> str:
        return "not interested"


class HaveAll(EmptyMessage):
    TYPE = 0x0E

    def __str__(self) -> str:
        return "have all"


class HaveNone(EmptyMessage):
    TYPE = 0x0F

    def __str__(self) -> str:
        return "have none"


@attr.define(slots=False)
class Have(Message):
    TYPE: ClassVar[int] = 4
    index: int

    def __str__(self) -> str:
        return f"have piece {self.index}"

    @classmethod
    def from_payload(cls, payload: bytes) -> Have:
        if len(payload) != 4:
            raise ValueError(
                f"Invalid length for 'have' payload; expected 4 bytes,"
                f" got {len(payload)}"
            )
        return cls(index=int.from_bytes(payload, "big"))

    def to_payload(self) -> bytes:
        return self.index.to_bytes(4, "big")


@attr.define(slots=False)
class Bitfield(Message):
    TYPE: ClassVar[int] = 5
    payload: bytes

    def __str__(self) -> str:
        return f"have {self.have_amount} pieces"

    @classmethod
    def from_payload(cls, payload: bytes) -> Bitfield:
        return cls(payload)

    def to_payload(self) -> bytes:
        return self.payload

    @property
    def have_amount(self) -> int:
        qty = 0
        for b in self.payload:
            for i in range(8):
                if b & (1 << i):
                    qty += 1
        return qty


@attr.define(slots=False)
class Request(Message):
    TYPE: ClassVar[int] = 6
    index: int
    begin: int
    length: int

    def __str__(self) -> str:
        return f"request piece {self.index}, offset {self.begin}, length {self.length}"

    @classmethod
    def from_payload(cls, payload: bytes) -> Request:
        if len(payload) != 12:
            raise ValueError(
                f"Invalid length for 'request' payload; expected 12 bytes,"
                f" got {len(payload)}"
            )
        index, begin, length = struct.unpack("!III", payload)
        return cls(index, begin, length)

    def to_payload(self) -> bytes:
        return struct.pack("!III", self.index, self.begin, self.length)


@attr.define(slots=False)
class Piece(Message):
    TYPE: ClassVar[int] = 7
    index: int
    begin: int
    data: bytes

    def __str__(self) -> str:
        return f"piece {self.index}, offset {self.begin}, length {len(self.data)}"

    @classmethod
    def from_payload(cls, payload: bytes) -> Piece:
        if len(payload) < 8:
            raise ValueError(
                f"Invalid length for 'piece' payload; expected 8+ bytes,"
                f" got {len(payload)}"
            )
        index, begin = struct.unpack_from("!II", payload)
        return cls(index, begin, payload[8:])

    def to_payload(self) -> bytes:
        return struct.pack("!II", self.index, self.begin) + self.data


@attr.define(slots=False)
class Cancel(Message):
    TYPE: ClassVar[int] = 8
    index: int
    begin: int
    length: int

    def __str__(self) -> str:
        return (
            f"cancel request for piece {self.index}, offset {self.begin},"
            f" length {self.length}"
        )

    @classmethod
    def from_payload(cls, payload: bytes) -> Cancel:
        if len(payload) != 12:
            raise ValueError(
                f"Invalid length for 'cancel' payload; expected 12 bytes,"
                f" got {len(payload)}"
            )
        index, begin, length = struct.unpack("!III", payload)
        return cls(index, begin, length)

    def to_payload(self) -> bytes:
        return struct.pack("!III", self.index, self.begin, self.length)


@attr.define(slots=False)
class Port(Message):
    TYPE: ClassVar[int] = 9
    port: int

    def __str__(self) -> str:
        return f"DHT port {self.port}"

    @classmethod
    def from_payload(cls, payload: bytes) -> Port:
        if len(payload) != 2:
            raise ValueError(
                f"Invalid length for 'port' payload; expected 2 bytes,"
                f" got {len(payload)}"
            )
        return cls(int.from_bytes(payload, "big"))

    def to_payload(self) -> bytes:
        return self.port.to_bytes(2, "big")


@attr.define(slots=False)
class Reject(Message):
    TYPE: ClassVar[int] = 0x10
    index: int
    begin: int
    length: int

    def __str__(self) -> str:
        return (
            f"reject request for piece {self.index}, offset {self.begin},"
            f" length {self.length}"
        )

    @classmethod
    def from_payload(cls, payload: bytes) -> Reject:
        if len(payload) != 12:
            raise ValueError(
                f"Invalid length for 'reject' payload; expected 12 bytes,"
                f" got {len(payload)}"
            )
        index, begin, length = struct.unpack("!III", payload)
        return cls(index, begin, length)

    def to_payload(self) -> bytes:
        return struct.pack("!III", self.index, self.begin, self.length)


@attr.define(slots=False)
class AllowedFast(Message):
    TYPE: ClassVar[int] = 0x11
    index: int

    def __str__(self) -> str:
        return f"allow fast download of piece {self.index}"

    @classmethod
    def from_payload(cls, payload: bytes) -> AllowedFast:
        if len(payload) != 4:
            raise ValueError(
                f"Invalid length for 'allowed fast' payload; expected 4 bytes,"
                f" got {len(payload)}"
            )
        index = int.from_bytes(payload, "big")
        return cls(index=index)

    def to_payload(self) -> bytes:
        return self.index.to_bytes(4, "big")


@attr.define(slots=False)
class Suggest(Message):
    TYPE: ClassVar[int] = 0x0D
    index: int

    def __str__(self) -> str:
        return f"suggest piece {self.index}"

    @classmethod
    def from_payload(cls, payload: bytes) -> Suggest:
        if len(payload) != 4:
            raise ValueError(
                f"Invalid length for 'suggest' payload; expected 4 bytes,"
                f" got {len(payload)}"
            )
        index = int.from_bytes(payload, "big")
        return cls(index=index)

    def to_payload(self) -> bytes:
        return self.index.to_bytes(4, "big")


@attr.define(slots=False)
class Extended(Message):
    TYPE: ClassVar[int] = 20
    msg_id: int
    payload: bytes

    def __str__(self) -> str:
        return f"extended message, ID {self.msg_id}"

    @classmethod
    def from_payload(cls, payload: bytes) -> Extended:
        if len(payload) < 1:
            raise ValueError(
                f"Invalid length for 'extended' payload; expected 1+ bytes,"
                f" got {len(payload)}"
            )
        return cls(payload[0], payload[1:])

    def to_payload(self) -> bytes:
        return bytes([self.msg_id]) + self.payload

    def decompose(
        self, extensions: BEP10Registry
    ) -> ExtendedHandshake | ExtendedMessage:
        if self.msg_id == 0:
            return ExtendedHandshake.from_extended_payload(self.payload)
        else:
            try:
                ext = extensions.from_code[self.msg_id]
            except KeyError:
                raise ValueError(f"Unknown extended message ID {self.msg_id}")
            for klass in ExtendedMessage.__subclasses__():
                if klass.EXTENSION == ext:
                    return klass.from_extended_payload(self.payload)
            raise ValueError(f"Unimplemented extended message ID {self.msg_id}")


@attr.define(slots=False)
class ExtendedHandshake:
    data: dict

    def __str__(self) -> str:
        s = "BEP 10 extended handshake; extensions: " + (
            ", ".join(self.extension_names) or "<none>"
        )
        if self.client is not None:
            s += f"; client: {self.client}"
        return s

    @classmethod
    def make(
        cls,
        extensions: BEP10Registry,
        client: Optional[str] = None,
        metadata_size: Optional[int] = None,
    ) -> ExtendedHandshake:
        data: dict[bytes, Any] = {b"m": extensions.to_m()}
        if client is not None:
            data[b"v"] = client.encode("utf-8")
        if metadata_size is not None:
            data[b"metadata_size"] = metadata_size
        return cls(data)

    @classmethod
    def from_extended_payload(cls, payload: bytes) -> ExtendedHandshake:
        try:
            data = unbencode(payload)
        except UnbencodeError as e:
            raise ValueError(f"invalid bencoded data: {e}")
        if not isinstance(data, dict):
            raise ValueError("payload is not a dict")
        if not isinstance(data.get(b"m"), dict):
            raise ValueError("'m' dictionary missing")
        return cls(data)

    def to_extended_payload(self) -> bytes:
        return bencode(self.data)

    def to_extended(self) -> Extended:
        return Extended(msg_id=0, payload=self.to_extended_payload())

    @property
    def extension_names(self) -> list[str]:
        exts: list[str] = []
        for k in self.data[b"m"]:
            assert isinstance(k, bytes)
            exts.append(k.decode("utf-8", "replace"))
        return exts

    @property
    def extensions(self) -> BEP10Registry:
        return BEP10Registry.from_m(self.data[b"m"])

    @property
    def client(self) -> Optional[str]:
        return get_string(self.data, b"v")

    @property
    def metadata_size(self) -> Optional[int]:
        return get_typed_value(self.data, b"metadata_size", int)


@attr.define(slots=False)
class ExtendedMessage(ABC):
    EXTENSION: ClassVar[BEP10Extension]

    @classmethod
    @abstractmethod
    def from_extended_payload(cls, payload: bytes) -> ExtendedMessage:
        ...

    @abstractmethod
    def to_extended_payload(self) -> bytes:
        ...

    def to_extended(self, extensions: BEP10Registry) -> Extended:
        msg_id = extensions.to_code[self.EXTENSION]
        return Extended(msg_id=msg_id, payload=self.to_extended_payload())


@attr.define(slots=False)
class BEP9Message(ExtendedMessage):
    EXTENSION: ClassVar[BEP10Extension] = BEP10Extension.METADATA

    msg_type: int
    piece: int
    total_size: Optional[int] = None
    payload: bytes = b""

    def __str__(self) -> str:
        try:
            typename = BEP9MsgType(self.msg_type).name
        except ValueError:
            typename = f"msg_type {self.msg_type}"
        return f"ut_metadata message ({typename} piece {self.piece})"

    @classmethod
    def from_extended_payload(cls, payload: bytes) -> BEP9Message:
        try:
            data, trailing = partial_unbencode(payload)
        except UnbencodeError as e:
            raise ValueError(
                f"ut_metadata message does not start with valid bencode: {e}"
            )
        if not isinstance(data, dict):
            raise ValueError("ut_metadata message does not start with a dict")
        if not isinstance(msg_type := data.get(b"msg_type"), int):
            raise ValueError("ut_metadata message lacks valid 'msg_type' field")
        if not isinstance(piece := data.get(b"piece"), int):
            raise ValueError("ut_metadata message lacks valid 'piece' field")
        total_size = data.get(b"total_size")
        if total_size is not None and not isinstance(total_size, int):
            raise ValueError("ut_metadata message has invalid 'total_size' field")
        try:
            mt = BEP9MsgType(msg_type)
        except ValueError:
            pass
        else:
            if mt is not BEP9MsgType.DATA:
                if trailing:
                    raise ValueError("Non-data ut_metadata message has trailing bytes")
            elif not trailing:
                raise ValueError("ut_metadata data message lacks trailing data")
        return cls(
            msg_type=msg_type,
            piece=piece,
            total_size=total_size,
            payload=trailing,
        )

    def to_extended_payload(self) -> bytes:
        data = {b"msg_type": self.msg_type, b"piece": self.piece}
        if self.total_size is not None:
            data[b"total_size"] = self.total_size
        return bencode(data) + self.payload


AnyMessage = Message | ExtendedHandshake | ExtendedMessage


def decode_message(blob: bytes, extensions: BEP10Registry) -> AnyMessage:
    msg = Message.parse(blob)
    if isinstance(msg, Extended):
        return msg.decompose(extensions)
    else:
        return msg


def encode_message(msg: AnyMessage, extensions: BEP10Registry) -> bytes:
    if isinstance(msg, ExtendedHandshake):
        msg = msg.to_extended()
    elif isinstance(msg, ExtendedMessage):
        msg = msg.to_extended(extensions)
    return bytes(msg)



"""
Convert magnet links to .torrent files

Visit <https://github.com/jwodder/demagnetize> for more information.
"""

__version__ = "0.2.0"
__author__ = "John Thorvald Wodder II"
__author_email__ = "demagnetize@varonathe.org"
__license__ = "MIT"
__url__ = "https://github.com/jwodder/demagnetize"



# <https://www.bittorrent.org/beps/bep_0015.html>
from __future__ import annotations
from collections.abc import Callable
from contextlib import AbstractContextManager, nullcontext
from functools import partial
from random import randint
from socket import AF_INET6
import struct
from time import time
from typing import TYPE_CHECKING, Any, ClassVar, Optional, TypeVar
from anyio import create_connected_udp_socket, fail_after
from anyio.abc import ConnectedUDPSocket, SocketAttribute
import attr
from .base import (
    AnnounceEvent,
    AnnounceResponse,
    Tracker,
    TrackerSession,
    unpack_peers,
    unpack_peers6,
)
from ..consts import LEFT, NUMWANT
from ..errors import TrackerFailure
from ..util import TRACE, InfoHash, Key, log

if TYPE_CHECKING:
    from ..core import Demagnetizer

T = TypeVar("T")

PROTOCOL_ID = 0x41727101980


@attr.define
class UDPTracker(Tracker):
    SCHEMES: ClassVar[list[str]] = ["udp"]

    host: str = attr.field(init=False)
    port: int = attr.field(init=False)

    def __attrs_post_init__(self) -> None:
        if self.url.scheme != "udp":
            raise ValueError("URL scheme must be 'udp'")
        if self.url.host is None:
            raise ValueError("URL missing host")
        if self.url.port is None:
            raise ValueError("URL missing port")
        self.host = self.url.host
        self.port = self.url.port

    async def connect(self, app: Demagnetizer) -> UDPTrackerSession:
        s = await create_connected_udp_socket(self.host, self.port)
        return UDPTrackerSession(tracker=self, app=app, socket=s)


@attr.define
class UDPTrackerSession(TrackerSession):
    tracker: UDPTracker
    app: Demagnetizer
    socket: ConnectedUDPSocket
    connection: Optional[Connection] = None

    @property
    def is_ipv6(self) -> bool:
        return self.socket.extra(SocketAttribute.family) == AF_INET6

    async def aclose(self) -> None:
        await self.socket.aclose()

    async def get_connection(self) -> Connection:
        if self.connection is None:
            log.log(TRACE, "Sending connection request to %s", self.tracker)
            transaction_id = make_transaction_id()
            conn_id = await self.send_receive(
                build_connection_request(transaction_id),
                partial(parse_connection_response, transaction_id),
            )
            self.connection = Connection(session=self, id=conn_id)
        return self.connection

    def reset_connection(self) -> None:
        self.connection = None

    async def announce(
        self,
        info_hash: InfoHash,
        event: AnnounceEvent,
        downloaded: int = 0,
        uploaded: int = 0,
        left: int = LEFT,
    ) -> UDPAnnounceResponse:
        while True:
            conn = await self.get_connection()
            try:
                return await conn.announce(
                    info_hash,
                    event,
                    downloaded=downloaded,
                    uploaded=uploaded,
                    left=left,
                    urldata=self.tracker.url.path_qs,
                )
            except ConnectionTimeoutError:
                log.log(
                    TRACE,
                    "Connection to %s timed out; restarting",
                    self.tracker,
                )
                self.reset_connection()

    async def send_receive(
        self,
        msg: bytes,
        response_parser: Callable[[bytes], T],
        expiration: Optional[float] = None,
    ) -> T:
        ctx: AbstractContextManager[Any]
        if expiration is None:
            ctx = nullcontext()
        else:
            ctx = fail_after(expiration - time())
        try:
            with ctx:
                n = 0
                while True:
                    await self.socket.send(msg)
                    try:
                        with fail_after(15 << n):
                            resp = await self.socket.receive()
                    except TimeoutError:
                        log.log(
                            TRACE,
                            "%s did not reply in time; resending message",
                            self.tracker,
                        )
                        if n < 8:
                            ### TODO: Should this count remember timeouts from
                            ### previous connections & connection attempts?
                            n += 1
                        continue
                    try:
                        data = response_parser(resp)
                    except TrackerFailure:
                        raise
                    except Exception as e:
                        log.log(TRACE, "Bad response from %s: %r", self.tracker, resp)
                        log.log(
                            TRACE,
                            "Response from %s was invalid, will resend: %s: %s",
                            self.tracker,
                            type(e).__name__,
                            e,
                        )
                        continue
                    return data
        except TimeoutError:
            raise ConnectionTimeoutError


@attr.define
class Connection:
    session: UDPTrackerSession
    id: int  # noqa: A003
    expiration: float = attr.field(init=False)

    def __attrs_post_init__(self) -> None:
        self.expiration = time() + 60

    async def announce(
        self,
        info_hash: InfoHash,
        event: AnnounceEvent,
        *,
        downloaded: int = 0,
        uploaded: int = 0,
        left: int = LEFT,
        urldata: str,
    ) -> UDPAnnounceResponse:
        transaction_id = make_transaction_id()
        msg = build_announce_request(
            transaction_id=transaction_id,
            connection_id=self.id,
            info_hash=info_hash,
            peer_id=self.session.app.peer_id,
            peer_port=self.session.app.peer_port,
            key=self.session.app.key,
            event=event,
            downloaded=downloaded,
            uploaded=uploaded,
            left=left,
            urldata=urldata,
        )
        return await self.session.send_receive(
            msg,
            partial(
                parse_announce_response,
                transaction_id,
                is_ipv6=self.session.is_ipv6,
            ),
            expiration=self.expiration,
        )


class ConnectionTimeoutError(Exception):
    pass


@attr.define
class UDPAnnounceResponse(AnnounceResponse):
    leechers: int
    seeders: int


def make_transaction_id() -> int:
    return randint(-(1 << 31), (1 << 31) - 1)


def build_connection_request(transaction_id: int) -> bytes:
    return struct.pack("!qii", PROTOCOL_ID, 0, transaction_id)


def raise_error_response(resp: bytes) -> None:
    action, _ = struct.unpack_from("!ii", resp)
    ### TODO: Should we ever care about checking the transaction ID?
    if action == 3:
        msg = resp[8:].decode("utf-8", "replace")
        raise TrackerFailure(msg)


def parse_connection_response(transaction_id: int, resp: bytes) -> int:
    # Returns connection ID or raises error
    raise_error_response(resp)
    action, xaction_id, connection_id = struct.unpack_from("!iiq", resp)
    # Use `struct.unpack_from()` instead of `unpack()` because "Clients ...
    # should not assume packets to be of a certain size"
    if xaction_id != transaction_id:
        raise ValueError(
            f"Transaction ID mismatch: expected {transaction_id}, got {xaction_id}"
        )
    if action != 0:
        raise ValueError(f"Action mismatch: expected 0, got {action}")
    assert isinstance(connection_id, int)
    return connection_id


def build_announce_request(
    *,
    transaction_id: int,
    connection_id: int,
    info_hash: InfoHash,
    peer_id: bytes,
    peer_port: int,
    key: Key,
    event: AnnounceEvent,
    downloaded: int,
    uploaded: int,
    left: int,
    urldata: str,
    numwant: int = NUMWANT,
) -> bytes:
    ip_address = b"\0\0\0\0"
    bs = (
        struct.pack("!qii", connection_id, 1, transaction_id)
        + bytes(info_hash)
        + (peer_id + b"\0" * 20)[:20]
        + struct.pack("!qqqi", downloaded, left, uploaded, event.udp_value)
        + ip_address
        + bytes(key)
        + struct.pack("!iH", numwant, peer_port)
    )
    # BEP 41
    ud = urldata.encode("utf-8")
    while ud:
        segment = ud[:255]
        ud = ud[255:]
        bs += bytes([0x02, len(segment)]) + segment
    return bs


def parse_announce_response(
    transaction_id: int, resp: bytes, is_ipv6: bool
) -> UDPAnnounceResponse:
    raise_error_response(resp)
    header = struct.Struct("!iiiii")
    action, xaction_id, interval, leechers, seeders = header.unpack_from(resp)
    if xaction_id != transaction_id:
        raise ValueError(
            f"Transaction ID mismatch: expected {transaction_id}, got {xaction_id}"
        )
    if action != 1:
        raise ValueError(f"Action mismatch: expected 1, got {action}")
    resp = resp[header.size :]
    if is_ipv6:
        peers = unpack_peers6(resp)
    else:
        peers = unpack_peers(resp)
    return UDPAnnounceResponse(
        interval=interval, leechers=leechers, seeders=seeders, peers=peers
    )



from __future__ import annotations
from base64 import b32decode
from collections.abc import AsyncIterator, Awaitable, Iterable, Iterator
from contextlib import AsyncExitStack, asynccontextmanager
from hashlib import sha1
import logging
from random import choices, randrange
import re
from string import ascii_letters, digits
from typing import Any, Optional, TypeVar, cast
from anyio import CapacityLimiter, create_memory_object_stream, create_task_group
from anyio.streams.memory import MemoryObjectSendStream
import attr
from torf import Magnet, Torrent
from .consts import INFO_CHUNK_SIZE, PEER_ID_PREFIX

log = logging.getLogger(__package__)

TRACE = 5

T = TypeVar("T")


@attr.define
class InfoHash:
    as_str: str = attr.field(eq=False)
    as_bytes: bytes

    @classmethod
    def from_string(cls, s: str) -> InfoHash:
        if len(s) == 40:
            b = bytes.fromhex(s)
        elif len(s) == 32:
            b = b32decode(s)
        else:
            raise ValueError(f"Invalid info hash: {s!r}")
        return cls(as_str=s, as_bytes=b)

    @classmethod
    def from_bytes(cls, b: bytes) -> InfoHash:
        if len(b) != 20:
            raise ValueError(f"Invalid info hash: {b!r}")
        return cls(as_str=b.hex(), as_bytes=b)

    def __str__(self) -> str:
        return self.as_str

    def __bytes__(self) -> bytes:
        return self.as_bytes

    @property
    def as_hex(self) -> str:
        return self.as_bytes.hex()


@attr.define
class Key:
    value: int

    @classmethod
    def generate(cls) -> Key:
        return cls(randrange(1 << 32))

    def __int__(self) -> int:
        return self.value

    def __str__(self) -> str:
        return f"{self.value:08x}"

    def __bytes__(self) -> bytes:
        return self.value.to_bytes(4, "big")


@attr.define
class Report:
    #: Collection of magnet links and the files their torrents were saved to
    #: (None if the demagnetization failed)
    downloads: list[tuple[Magnet, Optional[str]]] = attr.Factory(list)

    @classmethod
    def for_success(cls, magnet: Magnet, filename: str) -> Report:
        return cls(downloads=[(magnet, filename)])

    @classmethod
    def for_failure(cls, magnet: Magnet) -> Report:
        return cls(downloads=[(magnet, None)])

    @property
    def total(self) -> int:
        return len(self.downloads)

    @property
    def finished(self) -> int:
        return sum(1 for _, fname in self.downloads if fname is not None)

    @property
    def ok(self) -> bool:
        return bool(self.downloads) and all(
            fname is not None for _, fname in self.downloads
        )

    def __add__(self, other: Report) -> Report:
        return type(self)(self.downloads + other.downloads)

    def __iadd__(self, other: Report) -> Report:
        self.downloads.extend(other.downloads)
        return self


def yield_lines(fp: Iterable[str]) -> Iterator[str]:
    for line in fp:
        line = line.strip()
        if line and not line.startswith("#"):
            yield line


def template_torrent_filename(fntemplate: str, torrent: Torrent) -> str:
    fields = {
        "name": sanitize_pathname(str(torrent.name)),
        "hash": torrent.infohash,
    }
    return fntemplate.format_map(fields)


def sanitize_pathname(s: str) -> str:
    return re.sub(r'[\0-\x1F\x5C/<>:|"?*]', "_", re.sub(r"\s", " ", s))


def make_peer_id() -> bytes:
    s = PEER_ID_PREFIX.encode("utf-8")[:20]
    if len(s) < 20:
        s += "".join(choices(ascii_letters + digits, k=20 - len(s))).encode("us-ascii")
    return s


@asynccontextmanager
async def acollect(
    coros: Iterable[Awaitable[T]], limit: Optional[CapacityLimiter] = None
) -> AsyncIterator[AsyncIterator[T]]:
    async with create_task_group() as tg:
        sender, receiver = create_memory_object_stream()
        async with sender:
            for c in coros:
                tg.start_soon(_acollect_pipe, c, limit, sender.clone())
        async with receiver:
            yield receiver


async def _acollect_pipe(
    coro: Awaitable[T],
    limit: Optional[CapacityLimiter],
    sender: MemoryObjectSendStream[T],
) -> None:
    async with AsyncExitStack() as stack:
        if limit is not None:
            await stack.enter_async_context(limit)
        await stack.enter_async_context(sender)
        value = await coro
        await sender.send(value)


@attr.define
class InfoPiecer:
    total_size: int
    data: bytearray = attr.Factory(bytearray)
    sizes: list[int] = attr.field(init=False)
    index: int = 0
    digest: Any = attr.Factory(sha1)

    def __attrs_post_init__(self) -> None:
        qty, residue = divmod(self.total_size, INFO_CHUNK_SIZE)
        self.sizes = [INFO_CHUNK_SIZE] * qty
        if residue:
            self.sizes.append(residue)

    @property
    def piece_qty(self) -> int:
        return len(self.sizes)

    def add_piece(self, blob: bytes) -> None:
        if self.index >= len(self.sizes):
            raise ValueError("Too many pieces")
        expected_len = self.sizes[self.index]
        if len(blob) != expected_len:
            raise ValueError(
                f"Piece {self.index} is wrong length: expected {expected_len}"
                f" bytes, got {len(blob)}"
            )
        self.data.extend(blob)
        self.digest.update(blob)
        self.index += 1

    def get_data(self) -> bytes:
        return bytes(self.data)

    def get_digest(self) -> str:
        return cast(str, self.digest.hexdigest())


def get_typed_value(data: dict[bytes, Any], key: bytes, klass: type[T]) -> Optional[T]:
    if isinstance(value := data.get(key), klass):
        return value
    else:
        return None


def get_string(data: dict[bytes, Any], key: bytes) -> Optional[str]:
    if isinstance(value := data.get(key), bytes):
        return value.decode("utf-8", "replace")
    else:
        return None



from __future__ import annotations
from pathlib import Path
from random import randint
from time import time
from anyio import CapacityLimiter
import attr
import click
from torf import Magnet, Torrent
from torf._utils import decode_dict
from .consts import CLIENT, MAGNET_LIMIT
from .errors import DemagnetizeError
from .session import TorrentSession
from .util import Key, Report, acollect, log, make_peer_id, template_torrent_filename


@attr.define
class Demagnetizer:
    key: Key = attr.Factory(Key.generate)
    peer_id: bytes = attr.Factory(make_peer_id)
    peer_port: int = attr.Factory(lambda: randint(1025, 65535))

    def __attrs_post_init__(self) -> None:
        log.debug("Using key = %s", self.key)
        log.debug("Using peer ID = %r", self.peer_id)
        log.debug("Using peer port = %d", self.peer_port)

    async def download_torrent_info(
        self, magnets: list[Magnet], fntemplate: str
    ) -> Report:
        report = Report()
        coros = [self.demagnetize2file(m, fntemplate) for m in magnets]
        async with acollect(coros, limit=CapacityLimiter(MAGNET_LIMIT)) as ait:
            async for r in ait:
                report += r
        return report

    async def demagnetize2file(self, magnet: Magnet, fntemplate: str) -> Report:
        try:
            torrent = await self.demagnetize(magnet)
            filename = template_torrent_filename(fntemplate, torrent)
            log.info(
                "Saving torrent for info hash %s to file %s", magnet.infohash, filename
            )
        except DemagnetizeError as e:
            log.error("%s", e)
            return Report.for_failure(magnet)
        try:
            Path(filename).parent.mkdir(parents=True, exist_ok=True)
            with click.open_file(filename, "wb") as fp:
                torrent.write_stream(fp)
        except Exception as e:
            log.error(
                "Error writing torrent to file %r: %s: %s",
                filename,
                type(e).__name__,
                e,
            )
            return Report.for_failure(magnet)
        return Report.for_success(magnet, filename)

    async def demagnetize(self, magnet: Magnet) -> Torrent:
        session = self.open_session(magnet)
        md = await session.get_info()
        return compose_torrent(magnet, md)

    def open_session(self, magnet: Magnet) -> TorrentSession:
        return TorrentSession(app=self, magnet=magnet)


def compose_torrent(magnet: Magnet, info: dict) -> Torrent:
    torrent = Torrent()
    torrent.metainfo["info"] = decode_dict(info)
    torrent.trackers = magnet.tr
    torrent.created_by = CLIENT
    torrent.creation_date = time()
    return torrent



from __future__ import annotations
from typing import Optional
import pytest
from demagnetize.peer.extensions import (
    BEP9MsgType,
    BEP10Extension,
    BEP10Registry,
    Extension,
    extbit,
)
from demagnetize.peer.messages import (
    BEP9Message,
    Extended,
    ExtendedHandshake,
    Handshake,
    HaveAll,
    HaveNone,
    Message,
    Port,
)
from demagnetize.util import InfoHash


@pytest.mark.parametrize(
    "blob,handshake",
    [
        (
            b"\x13BitTorrent protocol\x00\x00\x00\x00\x00\x10\x00\x05k\xcb\xd4A"
            b"\xd7\xa0\x88\xc6;\xa8\xf8\x82\xe3\x12\x91\xd3\x85\xa7\x96L-TR3000"
            b"-vfu1svh0ewb6",
            Handshake(
                extensions={Extension.BEP10_EXTENSIONS, Extension.FAST, Extension.DHT},
                info_hash=InfoHash.from_bytes(
                    b"k\xcb\xd4A\xd7\xa0\x88\xc6;\xa8\xf8\x82\xe3\x12\x91\xd3"
                    b"\x85\xa7\x96L"
                ),
                peer_id=b"-TR3000-vfu1svh0ewb6",
            ),
        ),
        (
            b"\x13BitTorrent protocol\x00\x00\x00\x00\x00\x18\x00\x05k\xcb\xd4A"
            b"\xd7\xa0\x88\xc6;\xa8\xf8\x82\xe3\x12\x91\xd3\x85\xa7\x96L-qB4360"
            b"-5Ngjy9uIMl~O",
            Handshake(
                extensions={
                    Extension.BEP10_EXTENSIONS,
                    Extension.FAST,
                    Extension.DHT,
                    extbit(5, 0x08),
                },
                info_hash=InfoHash.from_bytes(
                    b"k\xcb\xd4A\xd7\xa0\x88\xc6;\xa8\xf8\x82\xe3\x12\x91\xd3"
                    b"\x85\xa7\x96L"
                ),
                peer_id=b"-qB4360-5Ngjy9uIMl~O",
            ),
        ),
    ],
)
def test_handshake(blob: bytes, handshake: Handshake) -> None:
    assert Handshake.parse(blob) == handshake
    assert bytes(handshake) == blob


@pytest.mark.parametrize(
    "blob,msg",
    [
        (b"\x00\x00\x00\x01\x0e", HaveAll()),
        (b"\x00\x00\x00\x01\x0f", HaveNone()),
        (b"\x00\x00\x00\x03\t\x88\xb7", Port(34999)),
        (b"\x00\x00\x00\x03\t\xea\xa5", Port(60069)),
        # Extended handshake:
        (
            b"\x00\x00\x00\xd5\x14\x00d12:complete_agoi1441e1:md11:lt_donthavei"
            b"7e10:share_modei8e11:upload_onlyi3e12:ut_holepunchi4e11:ut_metada"
            b"tai2e6:ut_pexi1ee13:metadata_sizei5436e4:reqqi500e11:upload_onlyi"
            b"1e1:v17:qBittorrent/4.3.66:yourip4:\x99\xa2D\x9be",
            Extended(
                msg_id=0,
                payload=(
                    b"d12:complete_agoi1441e1:md11:lt_donthavei7e10:share_modei"
                    b"8e11:upload_onlyi3e12:ut_holepunchi4e11:ut_metadatai2e6:u"
                    b"t_pexi1ee13:metadata_sizei5436e4:reqqi500e11:upload_onlyi"
                    b"1e1:v17:qBittorrent/4.3.66:yourip4:\x99\xa2D\x9be"
                ),
            ),
        ),
        # ut_pex:
        (
            b"\x00\x00\x00'\x14\x01d5:added12:V`\\\xe5\xc8\xd5\xb2\x9b\x8b\xa8"
            b"\x88\xb77:added.f2:\x10\x10e",
            Extended(
                msg_id=1,
                payload=(
                    b"d5:added12:V`\\\xe5\xc8\xd5\xb2\x9b\x8b\xa8\x88\xb77:adde"
                    b"d.f2:\x10\x10e"
                ),
            ),
        ),
        (
            b"\x00\x00\x00.\x14\x01d5:added18:V`\\\xe5\xc8\xd5\xb2\x9b\x8b\xa8"
            b"\x88\xb7\xb9\x15\xd8\x95\t\x847:added.f3:\x10\x12\x10e",
            Extended(
                msg_id=1,
                payload=(
                    b"d5:added18:V`\\\xe5\xc8\xd5\xb2\x9b\x8b\xa8\x88\xb7\xb9"
                    b"\x15\xd8\x95\t\x847:added.f3:\x10\x12\x10e"
                ),
            ),
        ),
        # ut_metadata:
        (
            b"\x00\x00\x00X\x14\x03d8:msg_typei1e5:piecei0e10:total_sizei5436ee"
            b"d5:filesld6:lengthi267661684e4:pathl72:...",
            Extended(
                msg_id=3,
                payload=(
                    b"d8:msg_typei1e5:piecei0e10:total_sizei5436eed5:filesld6:l"
                    b"engthi267661684e4:pathl72:..."
                ),
            ),
        ),
        (
            b"\x00\x00\x00\x1b\x14\x03d8:msg_typei0e5:piecei0ee",
            Extended(msg_id=3, payload=b"d8:msg_typei0e5:piecei0ee"),
        ),
    ],
)
def test_message(blob: bytes, msg: Message) -> None:
    assert Message.parse(blob) == msg
    assert bytes(msg) == blob


@pytest.mark.parametrize(
    "payload,handshake,extensions,extension_names,client,metadata_size",
    [
        (
            b"d12:complete_agoi1441e1:md11:lt_donthavei7e10:share_modei8e11:upl"
            b"oad_onlyi3e12:ut_holepunchi4e11:ut_metadatai2e6:ut_pexi1ee13:meta"
            b"data_sizei5436e4:reqqi500e11:upload_onlyi1e1:v17:qBittorrent/4.3."
            b"66:yourip4:\x99\xa2D\x9be",
            ExtendedHandshake(
                {
                    b"complete_ago": 1441,
                    b"m": {
                        b"lt_donthave": 7,
                        b"share_mode": 8,
                        b"upload_only": 3,
                        b"ut_holepunch": 4,
                        b"ut_metadata": 2,
                        b"ut_pex": 1,
                    },
                    b"metadata_size": 5436,
                    b"reqq": 500,
                    b"upload_only": 1,
                    b"v": b"qBittorrent/4.3.6",
                    b"yourip": b"\x99\xa2D\x9b",
                }
            ),
            BEP10Registry.from_dict(
                {BEP10Extension.METADATA: 2, BEP10Extension.PEX: 1}
            ),
            [
                "lt_donthave",
                "share_mode",
                "upload_only",
                "ut_holepunch",
                "ut_metadata",
                "ut_pex",
            ],
            "qBittorrent/4.3.6",
            5436,
        ),
        (
            b"d1:ei0e4:ipv616:\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\x011:md11:ut_metad"
            b"atai3e6:ut_pexi1ee1:pi60069e4:reqqi512e11:upload_onlyi0e1:v17:Tra"
            b"nsmission 3.00e",
            ExtendedHandshake(
                {
                    b"e": 0,
                    b"ipv6": b"\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\x01",
                    b"m": {
                        b"ut_metadata": 3,
                        b"ut_pex": 1,
                    },
                    b"p": 60069,
                    b"reqq": 512,
                    b"upload_only": 0,
                    b"v": b"Transmission 3.00",
                }
            ),
            BEP10Registry.from_dict(
                {BEP10Extension.METADATA: 3, BEP10Extension.PEX: 1}
            ),
            ["ut_metadata", "ut_pex"],
            "Transmission 3.00",
            None,
        ),
    ],
)
def test_extended_handshake(
    payload: bytes,
    handshake: ExtendedHandshake,
    extensions: BEP10Registry,
    extension_names: list[str],
    client: Optional[str],
    metadata_size: Optional[int],
) -> None:
    assert ExtendedHandshake.from_extended_payload(payload) == handshake
    assert handshake.to_extended() == Extended(0, payload)
    assert handshake.extensions == extensions
    assert handshake.extension_names == extension_names
    assert handshake.client == client
    assert handshake.metadata_size == metadata_size


@pytest.mark.parametrize(
    "ext,b9msg,msg_id",
    [
        (
            Extended(msg_id=3, payload=b"d8:msg_typei0e5:piecei0ee"),
            BEP9Message(msg_type=BEP9MsgType.REQUEST, piece=0),
            3,
        ),
        (
            Extended(
                msg_id=3,
                payload=(
                    b"d8:msg_typei1e5:piecei0e10:total_sizei5436eed5:filesld6:l"
                    b"engthi267661684e4:pathl72:..."
                ),
            ),
            BEP9Message(
                msg_type=BEP9MsgType.DATA,
                piece=0,
                total_size=5436,
                payload=b"d5:filesld6:lengthi267661684e4:pathl72:...",
            ),
            3,
        ),
    ],
)
def test_bep9_message(ext: Extended, b9msg: BEP9Message, msg_id: int) -> None:
    extensions = BEP10Registry.from_dict({BEP10Extension.METADATA: msg_id})
    assert ext.decompose(extensions) == b9msg
    assert b9msg.to_extended(extensions) == ext



from __future__ import annotations
from typing import TYPE_CHECKING, ClassVar, Optional, TypeVar, cast
from urllib.parse import quote
import attr
from httpx import AsyncClient, HTTPError
from .base import (
    AnnounceEvent,
    AnnounceResponse,
    Tracker,
    TrackerSession,
    unpack_peers,
    unpack_peers6,
)
from ..bencode import unbencode
from ..consts import CLIENT, LEFT, NUMWANT
from ..errors import TrackerError, TrackerFailure, UnbencodeError
from ..peer import Peer
from ..util import TRACE, InfoHash, get_string, get_typed_value, log

if TYPE_CHECKING:
    from ..core import Demagnetizer

T = TypeVar("T")


class HTTPTracker(Tracker):
    SCHEMES: ClassVar[list[str]] = ["http", "https"]

    async def connect(self, app: Demagnetizer) -> HTTPTrackerSession:
        return HTTPTrackerSession(
            tracker=self,
            app=app,
            client=AsyncClient(follow_redirects=True, headers={"User-Agent": CLIENT}),
        )


@attr.define
class HTTPTrackerSession(TrackerSession):
    tracker: HTTPTracker
    app: Demagnetizer
    client: AsyncClient

    async def aclose(self) -> None:
        await self.client.aclose()

    async def announce(
        self,
        info_hash: InfoHash,
        event: AnnounceEvent,
        downloaded: int = 0,
        uploaded: int = 0,
        left: int = LEFT,
    ) -> HTTPAnnounceResponse:
        # As of v0.22.0, the only way to send a bytes query parameter through
        # httpx is if we do all of the encoding ourselves.
        params = (
            f"info_hash={quote(bytes(info_hash))}"
            f"&peer_id={quote(self.app.peer_id)}"
            f"&port={self.app.peer_port}"
            f"&uploaded={uploaded}"
            f"&downloaded={downloaded}"
            f"&left={left}"
            f"&numwant={NUMWANT}"
            f"&key={quote(str(self.app.key))}"
            "&compact=1"
        )
        if event.http_value:
            params += f"&event={event.http_value}"
        url = self.tracker.url.with_fragment(None)
        if url.query_string:
            target = f"{url}&{params}"
        else:
            target = f"{url}?{params}"
        try:
            r = await self.client.get(target)
        except HTTPError as e:
            raise TrackerError(
                tracker=self.tracker,
                info_hash=info_hash,
                msg=f"{type(e).__name__}: {e}",
            )
        if r.is_error:
            raise TrackerError(
                tracker=self.tracker,
                info_hash=info_hash,
                msg=f"Request to tracker returned {r.status_code}",
            )
        try:
            response = HTTPAnnounceResponse.parse(r.content)
        except ValueError as e:
            log.log(TRACE, "Bad response from %s: %r", self.tracker, r.content)
            raise TrackerError(
                tracker=self.tracker, info_hash=info_hash, msg=f"Bad response: {e}"
            )
        if response.warning_message is not None:
            log.warning(
                "%s replied with warning: %s", self.tracker, response.warning_message
            )
        return response


@attr.define
class HTTPAnnounceResponse(AnnounceResponse):
    warning_message: Optional[str] = None
    min_interval: Optional[int] = None
    tracker_id: Optional[bytes] = None
    complete: Optional[int] = None
    incomplete: Optional[int] = None

    @classmethod
    def parse(cls, content: bytes) -> HTTPAnnounceResponse:
        # Unknown fields and (most) fields of the wrong type are discarded
        try:
            data = unbencode(content)
        except UnbencodeError as e:
            raise ValueError(f"invalid bencoded data: {e}")
        if not isinstance(data, dict):
            raise ValueError("invalid response")
        if (failure := data.get(b"failure reason")) is not None:
            if isinstance(failure, bytes):
                failure_reason = failure.decode("utf-8", "replace")
            else:
                # Do our best to salvage the situation
                failure_reason = str(failure)
            raise TrackerFailure(failure_reason)
        if (interval := get_typed_value(data, b"interval", int)) is None:
            # Just fill in a reasonable default
            interval = 1800
        peers: list[Peer] = []
        if b"peers" in data:
            if isinstance(data[b"peers"], list):
                # Original format (BEP 0003)
                for p in data[b"peers"]:
                    if not isinstance(p, dict):
                        raise ValueError("invalid 'peers' list")
                    try:
                        ip = cast(bytes, p[b"ip"]).decode("utf-8")
                    except Exception:
                        raise ValueError("invalid 'peers' list")
                    if b"port" in p and isinstance(p[b"port"], int):
                        port = p[b"port"]
                    else:
                        raise ValueError("invalid 'peers' list")
                    peer_id: Optional[bytes]
                    if b"peer id" in p and isinstance(p[b"peer id"], bytes):
                        peer_id = p[b"peer id"]
                    else:
                        peer_id = None
                    peers.append(Peer(host=ip, port=port, id=peer_id))
            elif isinstance(data[b"peers"], bytes):
                # Compact format (BEP 0023)
                peers.extend(unpack_peers(data[b"peers"]))
            else:
                raise ValueError("invalid 'peers' list")
        if b"peers6" in data:
            if not isinstance(data[b"peers6"], bytes):
                raise ValueError("invalid 'peers6' list")
            # Compact format (BEP 0007)
            peers.extend(unpack_peers6(data[b"peers6"]))
        return cls(
            interval=interval,
            peers=peers,
            warning_message=get_string(data, b"warning message"),
            min_interval=get_typed_value(data, b"min interval", int),
            tracker_id=get_typed_value(data, b"tracker id", bytes),
            complete=get_typed_value(data, b"complete", int),
            incomplete=get_typed_value(data, b"incomplete", int),
        )



from __future__ import annotations
import logging
import sys
from typing import Optional, TextIO
import anyio
import click
from click_loglevel import LogLevel
import colorlog
from torf import Magnet
from .core import Demagnetizer
from .util import TRACE, log, yield_lines


def validate_template(
    _ctx: click.Context, _param: click.Parameter, value: Optional[str]
) -> Optional[str]:
    if value is not None:
        try:
            value.format(name="name", hash="hash")
        except ValueError:
            raise click.BadParameter(f"{value}: invalid filename template")
    return value


@click.group()
@click.option(
    "-l",
    "--log-level",
    type=LogLevel(extra={"TRACE": TRACE}),
    default=logging.INFO,
    help="Set logging level  [default: INFO]",
)
def main(log_level: int) -> None:
    log.setLevel(log_level)
    colorlog.basicConfig(
        format="%(log_color)s%(asctime)s [%(levelname)-8s] %(name)s: %(message)s",
        datefmt="%H:%M:%S",
        log_colors={
            "TRACE": "green",
            "DEBUG": "cyan",
            "INFO": "bold",
            "WARNING": "yellow",
            "ERROR": "red",
            "CRITICAL": "bold_red",
        },
        level="INFO",
        stream=sys.stderr,
    )
    logging.addLevelName(TRACE, "TRACE")


@main.command()
@click.option("-o", "--outfile", default="{name}.torrent", callback=validate_template)
@click.argument("magnet", type=Magnet.from_string)
@click.pass_context
def get(ctx: click.Context, magnet: Magnet, outfile: str) -> None:
    """Convert a magnet link to a .torrent file"""
    demagnetizer = Demagnetizer()
    r = anyio.run(demagnetizer.demagnetize2file, magnet, outfile)
    if not r.ok:
        ctx.exit(1)


@main.command()
@click.option("-o", "--outfile", default="{name}.torrent", callback=validate_template)
@click.argument("magnetfile", type=click.File())
@click.pass_context
def batch(ctx: click.Context, magnetfile: TextIO, outfile: str) -> None:
    """Convert a collection of magnet links to .torrent files"""
    magnets: list[Magnet] = []
    ok = True
    with magnetfile:
        for line in yield_lines(magnetfile):
            try:
                m = Magnet.from_string(line)
            except ValueError:
                log.error("Invalid magnet link: %s", line)
                ok = False
            else:
                magnets.append(m)
    if not magnets:
        log.info("No magnet links to fetch")
        ctx.exit(0 if ok else 1)
    demagnetizer = Demagnetizer()
    r = anyio.run(demagnetizer.download_torrent_info, magnets, outfile)
    log.info(
        "%d/%d magnet links successfully converted to torrent files",
        r.finished,
        r.total,
    )
    if not r.ok:
        ctx.exit(1)


if __name__ == "__main__":
    main()



from . import __version__

#: Client string to use as the User-Agent for HTTP trackers, to send in
#: extended handshakes, and to use as the "Created by" field in Torrent files
CLIENT = f"demagnetize {__version__}"

#: Prefix for generated peer IDs
PEER_ID_PREFIX = "-DM-0020-"

#: "left" value to use when announcing to a tracker for a torrent we have only
#: the magnet link of
LEFT = 65535
# TODO: Look into appropriate values (For comparison, Transmission uses 2^63-1)

#: Number of peers to request per tracker
NUMWANT = 50

#: Extended message ID to declare for receiving BEP 9 messages
UT_METADATA = 42

#: Maximum length of a message to accept from a peer
MAX_PEER_MSG_LEN = 65535

#: Size of BEP 9 "data" message payloads
INFO_CHUNK_SIZE = 16 << 10

#: Overall timeout for interacting with a tracker
TRACKER_TIMEOUT = 30

#: Timeout for sending & receiving a "stopped" announcement to a tracker
TRACKER_STOP_TIMEOUT = 3

#: Maximum number of peers to interact with at once for a single magnet
PEERS_PER_MAGNET_LIMIT = 30

#: Timeout for connecting to a peer and performing the BitTorrent handshake
PEER_HANDSHAKE_TIMEOUT = 60

#: Maximum number of magnet links to operate on at once in batch mode
MAGNET_LIMIT = 50



from __future__ import annotations
from typing import TYPE_CHECKING
import attr

if TYPE_CHECKING:
    from .peer import Peer
    from .trackers import Tracker
    from .util import InfoHash


@attr.define
class TrackerError(Exception):
    tracker: Tracker
    info_hash: InfoHash
    msg: str

    def __str__(self) -> str:
        return f"Error announcing to {self.tracker} for {self.info_hash}: {self.msg}"


@attr.define
class PeerError(Exception):
    peer: Peer
    info_hash: InfoHash
    msg: str

    def __str__(self) -> str:
        return f"Error communicating with {self.peer} for {self.info_hash}: {self.msg}"


class DemagnetizeError(Exception):
    pass


class UnbencodeError(ValueError):
    pass


class TrackerFailure(Exception):
    # Raised when tracker returns a failure message
    ### TODO: Come up with a better name?
    pass



from __future__ import annotations
from enum import Enum, IntEnum
import attr


def extbit(bytenum: int, mask: int) -> int:
    for i in range(8):
        if 1 << i == mask:
            return i + 8 * (7 - bytenum)
    raise ValueError("Mask must be a single bit")


class Extension(IntEnum):
    AZUREUS_MESSAGING = extbit(0, 0x80)
    LOCATION_AWARE = extbit(2, 0x08)
    BEP10_EXTENSIONS = extbit(5, 0x10)
    DHT = extbit(7, 0x01)
    XBT_PEX = extbit(7, 0x02)
    FAST = extbit(7, 0x04)
    NAT_TRAVERSAL = extbit(7, 0x08)
    HYBRID_V2 = extbit(7, 0x10)


class BEP10Extension(Enum):
    METADATA = "ut_metadata"
    PEX = "ut_pex"


@attr.define
class BEP10Registry:
    from_code: dict[int, BEP10Extension] = attr.Factory(dict)
    to_code: dict[BEP10Extension, int] = attr.Factory(dict)

    def __contains__(self, ext: BEP10Extension) -> bool:
        return ext in self.to_code

    @classmethod
    def from_dict(cls, d: dict[BEP10Extension, int]) -> BEP10Registry:
        msg_ids = cls()
        for ext, code in d.items():
            msg_ids.register(ext, code)
        return msg_ids

    @classmethod
    def from_m(cls, m: dict) -> BEP10Registry:
        msg_ids = cls()
        for k, v in m.items():
            if not isinstance(k, bytes):
                continue
            try:
                bext = BEP10Extension(k.decode("utf-8", "replace"))
            except ValueError:
                continue
            if isinstance(v, int):
                msg_ids.register(bext, v)
        return msg_ids

    def register(self, ext: BEP10Extension, code: int) -> None:
        if code in self.from_code:
            raise ValueError(f"conflicting declarations for code {code}")
        if ext in self.to_code:
            raise ValueError(f"conflicting declarations for key {ext.value!r}")
        self.from_code[code] = ext
        self.to_code[ext] = code

    def to_m(self) -> dict[bytes, int]:
        return {k.value.encode("utf-8"): v for k, v in self.to_code.items()}


class BEP9MsgType(IntEnum):
    REQUEST = 0
    DATA = 1
    REJECT = 2
